import argparse
from daphne.management.commands.runserver import Command as DaphneCommand
from asgi_autoreload import AutoreloadMiddleware


class Command(DaphneCommand):
    help = "Run Daphne with ASGI autoreload browser injection"

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--watch",
            nargs="*",
            default=["."],
            dest="watch_paths",
            help="Paths to watch for changes (default: .)",
        )
        parser.add_argument(
            "--exclude",
            nargs="*",
            default=[],
            dest="exclude_paths",
            help="Paths to exclude from watching.",
        )
        parser.add_argument(
            "build_cmd",
            nargs=argparse.REMAINDER,
            help="Optional build command to run on startup and on each change (after --).",
        )

    def get_application(self, options):
        app = super().get_application(options)
        watch_paths = options.get("watch_paths") or ["."]
        build_cmd = options.get("build_cmd") or []
        # argparse REMAINDER includes the leading "--" separator if present
        if build_cmd and build_cmd[0] == "--":
            build_cmd = build_cmd[1:]
        exclude_paths = options.get("exclude_paths") or []
        return AutoreloadMiddleware(
            app,
            watch_paths=watch_paths,
            build_command=build_cmd or None,
            exclude_paths=exclude_paths or None,
        )



from fastapi import FastAPI, Request
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

app = FastAPI()
app.mount("/static", StaticFiles(directory="static"), name="static")
templates = Jinja2Templates(directory="templates")


@app.get("/", response_class=HTMLResponse)
async def index(request: Request):
    return templates.TemplateResponse(name="index.html", request=request)



from __future__ import annotations
import asyncio
from typing import Any
from ._script import SSE_SCRIPT
from .broadcaster import Broadcaster
from .watcher import FileWatcher


class AutoreloadMiddleware:
    def __init__(
        self,
        app: Any,
        watch_paths: list[str] = ["."],
        sse_path: str = "/_autoreload",
        build_command: list[str] | None = None,
        exclude_paths: list[str] | None = None,
    ) -> None:
        self._app = app
        self._watch_paths = watch_paths
        self._sse_path = sse_path
        self._broadcaster = Broadcaster()
        self._watcher = FileWatcher(
            watch_paths, self._broadcaster, build_command, exclude_paths
        )
        self._watcher_started = False

    async def __call__(self, scope: dict, receive: Any, send: Any) -> None:
        if scope["type"] == "lifespan":
            await self._handle_lifespan(scope, receive, send)
        elif scope["type"] == "http" and scope["path"] == self._sse_path:
            await self._handle_sse(scope, receive, send)
        elif scope["type"] == "http":
            await self._handle_http(scope, receive, send)
        else:
            await self._app(scope, receive, send)

    async def _handle_lifespan(self, scope: dict, receive: Any, send: Any) -> None:
        async def wrapped_receive() -> dict:
            message = await receive()
            if message["type"] == "lifespan.startup":
                self._start_watcher()
            elif message["type"] == "lifespan.shutdown":
                await self._watcher.stop()
            return message

        await self._app(scope, wrapped_receive, send)

    def _start_watcher(self) -> None:
        if not self._watcher_started:
            self._watcher.start()
            self._watcher_started = True

    async def _handle_sse(self, scope: dict, receive: Any, send: Any) -> None:
        self._start_watcher()
        queue = await self._broadcaster.subscribe()

        await send({
            "type": "http.response.start",
            "status": 200,
            "headers": [
                [b"content-type", b"text/event-stream"],
                [b"cache-control", b"no-cache"],
                [b"connection", b"keep-alive"],
            ],
        })

        async def sse_task() -> None:
            while True:
                event = await queue.get()
                data = f"event: {event}\ndata: \n\n".encode()
                await send({"type": "http.response.body", "body": data, "more_body": True})

        async def disconnect_task() -> None:
            while True:
                msg = await receive()
                if msg["type"] == "http.disconnect":
                    return

        t1 = asyncio.ensure_future(sse_task())
        t2 = asyncio.ensure_future(disconnect_task())
        try:
            done, pending = await asyncio.wait(
                {t1, t2}, return_when=asyncio.FIRST_COMPLETED
            )
            for t in pending:
                t.cancel()
        finally:
            await self._broadcaster.unsubscribe(queue)
            await send({"type": "http.response.body", "body": b"", "more_body": False})

    async def _handle_http(self, scope: dict, receive: Any, send: Any) -> None:
        interceptor = _ResponseInterceptor(send)
        await self._app(scope, receive, interceptor)
        await interceptor.flush()


class _ResponseInterceptor:
    def __init__(self, send: Any) -> None:
        self._send = send
        self._start: dict | None = None
        self._body = b""

    async def __call__(self, message: dict) -> None:
        if message["type"] == "http.response.start":
            self._start = message
        elif message["type"] == "http.response.body":
            self._body += message.get("body", b"")
        else:
            await self._send(message)

    async def flush(self) -> None:
        if self._start is None:
            return

        headers: list[list[bytes]] = list(self._start.get("headers", []))
        content_type = b""
        content_encoding = b""
        for name, value in headers:
            if name.lower() == b"content-type":
                content_type = value
            if name.lower() == b"content-encoding":
                content_encoding = value

        body = self._body
        if b"text/html" in content_type and not content_encoding:
            if b"</body>" in body:
                body = body.replace(b"</body>", SSE_SCRIPT + b"</body>", 1)
            else:
                body = body + SSE_SCRIPT
            headers = [
                [n, v] for n, v in headers if n.lower() != b"content-length"
            ]
            headers.append([b"content-length", str(len(body)).encode()])

        await self._send({**self._start, "headers": headers})
        await self._send({"type": "http.response.body", "body": body, "more_body": False})



import asyncio
import pytest
from unittest.mock import patch
from asgi_autoreload.middleware import AutoreloadMiddleware


# ── helpers ──────────────────────────────────────────────────────────────────

def html_app(content: bytes = b"<html><body>hello</body></html>"):
    async def app(scope, receive, send):
        await send({
            "type": "http.response.start",
            "status": 200,
            "headers": [[b"content-type", b"text/html; charset=utf-8"],
                        [b"content-length", str(len(content)).encode()]],
        })
        await send({"type": "http.response.body", "body": content, "more_body": False})
    return app


def plain_app(content: bytes = b"hello"):
    async def app(scope, receive, send):
        await send({
            "type": "http.response.start",
            "status": 200,
            "headers": [[b"content-type", b"text/plain"],
                        [b"content-length", str(len(content)).encode()]],
        })
        await send({"type": "http.response.body", "body": content, "more_body": False})
    return app


async def http_scope(path: str = "/") -> dict:
    return {"type": "http", "path": path, "query_string": b""}


async def fake_receive():
    return {"type": "http.disconnect"}


async def collect_response(middleware, scope):
    messages = []
    async def send(msg):
        messages.append(msg)
    await middleware(scope, fake_receive, send)
    return messages


# ── HTML injection ────────────────────────────────────────────────────────────

async def test_injects_script_before_body_close():
    mw = AutoreloadMiddleware(html_app())
    scope = await http_scope()
    msgs = await collect_response(mw, scope)
    body = msgs[-1]["body"]
    assert b"EventSource" in body
    assert body.endswith(b"</body></html>")


async def test_no_injection_for_plain_text():
    mw = AutoreloadMiddleware(plain_app())
    scope = await http_scope()
    msgs = await collect_response(mw, scope)
    body = msgs[-1]["body"]
    assert b"EventSource" not in body


async def test_content_length_updated_after_injection():
    mw = AutoreloadMiddleware(html_app())
    scope = await http_scope()
    msgs = await collect_response(mw, scope)
    headers = dict(msgs[0]["headers"])
    assert int(headers[b"content-length"]) == len(msgs[-1]["body"])


async def test_no_injection_when_content_encoding_present():
    async def compressed_app(scope, receive, send):
        await send({
            "type": "http.response.start",
            "status": 200,
            "headers": [[b"content-type", b"text/html"],
                        [b"content-encoding", b"gzip"]],
        })
        await send({"type": "http.response.body", "body": b"compressed", "more_body": False})

    mw = AutoreloadMiddleware(compressed_app)
    scope = await http_scope()
    msgs = await collect_response(mw, scope)
    assert b"EventSource" not in msgs[-1]["body"]


async def test_appends_script_when_no_body_tag():
    content = b"<html>no body tag</html>"
    mw = AutoreloadMiddleware(html_app(content))
    scope = await http_scope()
    msgs = await collect_response(mw, scope)
    body = msgs[-1]["body"]
    assert body.endswith(b"</script>")
    assert b"EventSource" in body


# ── SSE endpoint ──────────────────────────────────────────────────────────────

async def test_sse_endpoint_returns_200():
    mw = AutoreloadMiddleware(plain_app())
    scope = {"type": "http", "path": "/_autoreload", "query_string": b""}

    messages = []
    call_count = 0

    async def recv():
        nonlocal call_count
        call_count += 1
        if call_count == 1:
            await asyncio.sleep(0.01)
        return {"type": "http.disconnect"}

    async def send(msg):
        messages.append(msg)

    await mw(scope, recv, send)
    assert messages[0]["status"] == 200
    assert messages[0]["headers"][0][1] == b"text/event-stream"


async def test_sse_delivers_broadcast():
    mw = AutoreloadMiddleware(plain_app())
    scope = {"type": "http", "path": "/_autoreload", "query_string": b""}

    messages = []
    disconnected = asyncio.Event()

    async def recv():
        await disconnected.wait()
        return {"type": "http.disconnect"}

    async def send(msg):
        messages.append(msg)
        if msg.get("type") == "http.response.body" and msg.get("more_body"):
            disconnected.set()

    task = asyncio.ensure_future(mw(scope, recv, send))
    await asyncio.sleep(0.01)
    await mw._broadcaster.broadcast("reload")
    await task

    bodies = [m["body"] for m in messages if m.get("type") == "http.response.body" and m.get("body")]
    assert any(b"reload" in b for b in bodies)



import asyncio


class Broadcaster:
    def __init__(self) -> None:
        self._queues: set[asyncio.Queue[str]] = set()
        self._lock = asyncio.Lock()

    async def subscribe(self) -> asyncio.Queue[str]:
        q: asyncio.Queue[str] = asyncio.Queue()
        async with self._lock:
            self._queues.add(q)
        return q

    async def unsubscribe(self, q: asyncio.Queue[str]) -> None:
        async with self._lock:
            self._queues.discard(q)

    async def broadcast(self, event: str) -> None:
        async with self._lock:
            queues = set(self._queues)
        for q in queues:
            await q.put(event)



from fastapi import FastAPI, Request
from fastapi.responses import HTMLResponse
from fastapi.templating import Jinja2Templates

app = FastAPI()
templates = Jinja2Templates(directory="templates")


@app.get("/", response_class=HTMLResponse)
async def index(request: Request):
    return templates.TemplateResponse(request=request, name="index.html")


@app.get("/about", response_class=HTMLResponse)
async def about(request: Request):
    return templates.TemplateResponse(request=request, name="about.html")



from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent.parent

SECRET_KEY = "django-insecure-example-key-do-not-use-in-production"

DEBUG = True

ALLOWED_HOSTS = ["*"]

INSTALLED_APPS = [
    "django.contrib.staticfiles",
    "asgi_autoreload.django",
    "mysite",
]

ROOT_URLCONF = "mysite.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [BASE_DIR / "mysite" / "templates"],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.request",
            ],
        },
    },
]

ASGI_APPLICATION = "mysite.asgi.application"

DATABASES = {}

STATIC_URL = "static/"



from __future__ import annotations
import importlib
import sys

import click
import typer
import uvicorn
from fastapi_cli.cli import app as _fastapi_typer_app

from asgi_autoreload import AutoreloadMiddleware

_fastapi_click = typer.main.get_command(_fastapi_typer_app)
_original_dev = _fastapi_click.commands["dev"]

params = list(_original_dev.params) + [
    click.Option(
        ["--watch"],
        multiple=True,
        default=["."],
        show_default=True,
        help="Paths to watch for browser reload.",
    ),
    click.Option(
        ["--exclude"],
        multiple=True,
        default=[],
        show_default=True,
        help="Paths to exclude from watching.",
    ),
]

# Remove the "reload" option, because we always want to enable autoreload
params = [param for param in params if param.name != "reload"]


def make_patched_run(
    build_command: list[str] | None,
    original_run: callable,
    watch: tuple[str, ...],
    exclude: tuple[str, ...],
) -> callable:
    def patched_run(app, **run_kwargs):  # type: ignore[no-untyped-def]
        if isinstance(app, str):
            if "" not in sys.path:
                sys.path.insert(0, "")
            mod_str, attr = app.rsplit(":", 1)
            app = getattr(importlib.import_module(mod_str), attr)
        wrapped = AutoreloadMiddleware(
            app,
            watch_paths=list(watch),
            build_command=build_command,
            exclude_paths=list(exclude) or None,
        )
        run_kwargs.pop("reload", None)  # incompatible with Python object
        original_run(wrapped, **run_kwargs)

    return patched_run


@click.pass_context
def callback(
    ctx: click.Context, watch: tuple[str, ...], exclude: tuple[str, ...], **kwargs
) -> None:
    build_command = ctx.args if ctx.args else None
    original_run = uvicorn.run

    uvicorn.run = make_patched_run(build_command, original_run, watch, exclude)
    try:
        ctx.invoke(_original_dev, **kwargs)
    finally:
        uvicorn.run = original_run


dev_command = click.Command(
    name="dev",
    callback=callback,
    params=params,
    help=_original_dev.help,
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)


@click.group()
def main() -> None:
    pass


main.add_command(dev_command)



import asyncio
import sys
import pytest
from unittest.mock import patch
from asgi_autoreload.broadcaster import Broadcaster
from asgi_autoreload.watcher import FileWatcher


async def _fake_awatch(*paths, stop_event, **kwargs):
    await asyncio.sleep(0.01)
    yield {("modified", "/fake/file.py")}
    await stop_event.wait()


async def test_watcher_broadcasts_on_change():
    b = Broadcaster()
    q = await b.subscribe()
    w = FileWatcher(["."], b)

    with patch("asgi_autoreload.watcher.awatch", _fake_awatch):
        w.start()
        await asyncio.sleep(0.05)
        await w.stop()

    assert not q.empty()
    assert q.get_nowait() == "reload"


async def test_watcher_stop_before_change():
    b = Broadcaster()
    q = await b.subscribe()

    async def slow_awatch(*paths, stop_event, **kwargs):
        await stop_event.wait()
        return
        yield  # make it an async generator

    w = FileWatcher(["."], b)
    with patch("asgi_autoreload.watcher.awatch", slow_awatch):
        w.start()
        await w.stop()

    assert q.empty()


async def test_build_command_runs_on_startup():
    b = Broadcaster()
    ran = []

    async def slow_awatch(*paths, stop_event, **kwargs):
        await stop_event.wait()
        return
        yield

    w = FileWatcher(["."], b, build_command=[sys.executable, "-c", "import sys; sys.exit(0)"])
    with patch("asgi_autoreload.watcher.awatch", slow_awatch):
        w.start()
        await asyncio.sleep(0.05)
        await w.stop()
    # no exception = command ran successfully


async def test_build_command_runs_before_each_broadcast():
    b = Broadcaster()
    q = await b.subscribe()
    order: list[str] = []

    original_exec = asyncio.create_subprocess_exec

    async def fake_exec(*args, **kwargs):
        order.append("build")
        proc = await original_exec(*args, **kwargs)
        return proc

    w = FileWatcher(["."], b, build_command=[sys.executable, "-c", ""])
    with patch("asgi_autoreload.watcher.awatch", _fake_awatch), \
         patch("asyncio.create_subprocess_exec", fake_exec):
        w.start()
        await asyncio.sleep(0.05)
        await w.stop()

    # build ran on startup + on file change
    assert order.count("build") >= 2
    assert not q.empty()


async def test_excluded_path_does_not_trigger_reload(tmp_path):
    b = Broadcaster()
    q = await b.subscribe()
    excluded = tmp_path / "static"
    excluded.mkdir()

    async def fake_awatch(*paths, watch_filter, stop_event, **kwargs):
        # emit a change inside the excluded directory
        path = str(excluded / "bundle.css")
        if watch_filter(None, path):
            yield {("modified", path)}
        await stop_event.wait()

    w = FileWatcher([str(tmp_path)], b, exclude_paths=[str(excluded)])
    with patch("asgi_autoreload.watcher.awatch", fake_awatch):
        w.start()
        await asyncio.sleep(0.05)
        await w.stop()

    assert q.empty()


async def test_non_excluded_path_still_triggers_reload(tmp_path):
    b = Broadcaster()
    q = await b.subscribe()
    excluded = tmp_path / "static"
    excluded.mkdir()
    src = tmp_path / "src"
    src.mkdir()

    async def fake_awatch(*paths, watch_filter, stop_event, **kwargs):
        path = str(src / "style.css")
        if watch_filter(None, path):
            yield {("modified", path)}
        await stop_event.wait()

    w = FileWatcher([str(tmp_path)], b, exclude_paths=[str(excluded)])
    with patch("asgi_autoreload.watcher.awatch", fake_awatch):
        w.start()
        await asyncio.sleep(0.05)
        await w.stop()

    assert not q.empty()
